/**
 * Personal Compass
 * B
 * NOTE: CODE MUST BE COMPILED USING ARDUINO IDE version 0023
 * old IDEs available here: http://arduino.cc/en/Main/OldSoftwareReleases
 *
 * The user must supply a microphone preamp that delivers 0 - vref volts
 * of audio signal to the analog pin.  The preamp should deliver vref/2
 * volts for silence.
 *
 */





#include <TinyGPS.h> // download library from arduiniana.org
#include <NewSoftSerial.h> // download library from arduiniana.org ["SoftwareSerial" = new name for NewSoftSerial v11 beta]
#include <Wire.h> // built-in library; allows Arduino to communicate with I2C / TWI devices
#include <PWMServo.h> //new library doesn't interfere with PWM; plays nicely with NewSoftSerial/GPS

// GPS Info and test coordinates
#define EARTH_RADIUS 6371.009f 

// Tompkins Square dog park, NYC
#define DESTINATION_LATITUDE ( 40.726836f * DEG_TO_RAD )
#define DESTINATION_LONGITUDE (-73.980732f * DEG_TO_RAD )

// Pins
#define GPS_RX 2 // GPS recieve
#define GPS_TX 3 // GPS transmit
#define SDA 4 // analog input from compass
#define SCL 5 // analog input from compass
#define servoPointerPin 10 // servo

// Controller variables
TinyGPS gps;
NewSoftSerial nss( GPS_RX, GPS_TX ); 
PWMServo servoPointer;  
unsigned long lastgps;

// Variables for GPS calculations
float g_distanceToDestination = -1.0f; 
float g_directionToDestination = 0.0f;      
int g_currentHeading = 500;                
float flat, flon;
unsigned long fix_age;

// initializes servo, moves to servo value 10, then detaches servo (will be attached again once GPS lock attained)
void setupServo() {  
  servoPointer.attach( servoPointerPin );      
  positionServo( &servoPointer, 10 ); 
  servoPointer.detach(); 
}                                                

// initializes NewSoftSerial library at 4800 baud
void setupGps(){
    nss.begin(4800);
}

// starts wire library - initializes compass via analog serial inputs
void setupCompass(){
  Wire.begin();
}

void setup()
{
  Serial.begin(115200); // Initializes arduino built-in serial library; used only for serial debugging
  setupServo(); 
  delay(1000);  
  setupGpsStatus();   // sets up i/o pins and states for status LEDs
  setupCompass(); 
  setupGps();         
  Serial.println("");
  Serial.println("GPS Compass v. 1.0 starting up.");
  Serial.println(" ...waiting for GPS lock... ");
  Serial.println("");
  delay(10000); // gives gps/compass time to initialize before moving on to main program loop
}

//calculates compass bearing to a lat/lon point
float bearingTo( const float lat1, const float lon1, const float lat2, const float lon2 )
{
  const float dLat = lat2 - lat1;
  const float dLon = lon2 - lon1;
  const float y = sin( dLon ) * cos( lat2 );
  const float x = cos( lat1 ) * sin( lat2 ) - sin( lat1 ) * cos( lat2 ) * cos( dLon );
  float bearing = atan2(y, x) / DEG_TO_RAD;
  if( bearing < 0 )
    bearing = 360 + bearing;
  return bearing;
}


//calculates distance between two lat/lon points
float distanceBetween( const float lat1, const float lon1, const float lat2, const float lon2 )
{
  const float dLat = lat2 - lat1;
  const float dLon = lon2 - lon1;
  const float a = sin( dLat / 2.0 ) * sin( dLat / 2.0 ) + cos( lat1 ) * cos( lat2 ) * sin( dLon / 2.0 ) * sin( dLon / 2.0 );
  const float c = 2 * atan2( sqrt( a ),  sqrt( 1.0 - a ) ); 
  return fabs( EARTH_RADIUS * c );
}

// reads servo position, moves servo if relative bearing to GPS coordinate changes
void positionServo( PWMServo* pServo, int position ) {
  int currentPosition = pServo->read();
  if( currentPosition != position )
    pServo->write( position );
}



/////////////////////////////////////////////////
// Get distance and direction to target from GPS
////////////////////////////////////////////////

int lastBearing = -1;

bool getDistanceAndDirection( const float targetLat, const float targetLon, float& distanceToDestination, float& directionToDestination )
{
  nss.flush();
  unsigned long start = millis(); 
  while( ( ( millis() - start ) < 250 ) && ( millis() >= start ) )  // in first 250 ms of program...
  {
    if(nss.available() ) // ...if GPS is ready....
    {
      int c = nss.read(); // ...read the NSS ports...
      if( gps.encode( c ) )
      {
        unsigned long fix_age;
        float flat, flon;
        gps.f_get_position( &flat, &flon, &fix_age ); // ...and get coordinates for your current location.
        float currentLatitude = flat * DEG_TO_RAD; 
        float currentLongitude = flon * DEG_TO_RAD;

        //uncomment below for debugging; prints current lat/lon.
    /*
        Serial.print("Current Lat: ");
        Serial.println(flat,7);  // prints current latitude to 7th decimal pt.
        Serial.print("Current Lon: ");
        Serial.println(flon,7); // prints current longitude to 7th decimal pt.
    */
        float newDistanceToDestination = distanceBetween( currentLatitude, currentLongitude, targetLat, targetLon ) * 1000.0f;      
        float newDirectionToDestination = bearingTo( currentLatitude, currentLongitude, targetLat, targetLon );
        bool resultChanged = ( lastBearing == -1 ) ? true : ( abs( newDirectionToDestination - lastBearing ) > 2 );
        distanceToDestination = newDistanceToDestination;
        directionToDestination = newDirectionToDestination;
        if( resultChanged )
          lastBearing = directionToDestination;
        return resultChanged;

      }
    }
  }
  return false; 
}


/////////////////////////
// Read compass heading
/////////////////////////

int lastHeading = -1;

bool getCurrentHeading( int& reading ) // this code needed to interface with compass chip - shifts bits over by one
{
  int compassAddress = 0x42 >> 1;
  Wire.beginTransmission( compassAddress );
  Wire.send( 'A' );
  Wire.endTransmission();
  Wire.requestFrom( compassAddress, 2 );

  delay(10);  
  Wire.requestFrom( compassAddress, 2 );
  while( Wire.available() < 2 )
    delay(10);

  int newReading = Wire.receive();
  newReading = newReading << 8;
  newReading += Wire.receive();
  newReading /= 10;
  bool result = ( lastHeading == -1 ) ? true : ( abs( newReading - lastHeading ) > 2 );
  reading = newReading;
  if( result )
    lastHeading = reading;  
  return result;
}


// returns GPS ready to read

unsigned long lastGPSReading = -1;




bool ReadyToReadGPS()
{
  unsigned long now = millis();  
  
  if ( lastGPSReading == -1 ) // If GPS hasn't yet initialized (still outputs initial variable value)...
    return true;
  if( g_distanceToDestination == -1.0f || now < lastGPSReading ) 
    return true;  
  unsigned long ellapsed = now - lastGPSReading;  
  if( g_distanceToDestination < 500 ) // if distance to destination is smaller than 500 (not too far away)...
    return ellapsed > 1000;  // check the gps location more often (once/sec)...
  else
    return ellapsed > 10000; // otherwise, don't check it very often (once every 10 secs).
}


////////////////////////////////////////////////
// Test if GPS is recieving data from satellites
////////////////////////////////////////////////

bool feedgps()
{
  while (nss.available())
  {
    if (gps.encode(nss.read()))
      return true;
  }
  return false;
}


//////////////////////////////
//    PROGRAM LOOP  
//////////////////////////////

void loop()
{
  bool gpsDataChanged = false;

  if(ReadyToReadGPS())
    
  {
    gpsDataChanged = getDistanceAndDirection( DESTINATION_LATITUDE, DESTINATION_LONGITUDE, g_distanceToDestination, g_directionToDestination );
    lastGPSReading = millis(); 
  }

  // Compass and servo commands: 
  bool headingDataChanged = getCurrentHeading( g_currentHeading );  

  if( DataValid() && ( gpsDataChanged || headingDataChanged ) )
   PositionNeedle( g_distanceToDestination, g_directionToDestination, g_currentHeading );
 
  gpsServoLedSwitch(); // controls gps status lights. Also detaches servo if no GPS lock.

}


////////////////
// SERVO CONTROL
// change min/max values on servo sweep to match gears installed on final unit
///////////////

void PositionNeedle( const int distanceToDestination, const int directionToDestination, const int currentHeading )
{

  int bearing = 0;  // sweep
  if( currentHeading > directionToDestination ) 
    bearing = -(currentHeading - directionToDestination); // These two are equal. -(a-b) = -a+b
  else
    bearing = directionToDestination - currentHeading;  
    
   
  if( bearing < 0 ) 
    bearing += 360; // makes sure bearing is positive number; if smaller than zero, adds 360. Makes needle flip around when servo hits is rotation limit, giving illusion o 360 degree rotation

    //change values below to match gears on final product
  float servomin = 20.0f;
  float servomax = 150.0f;
  int servoPosition = (int)(servomax - (servomax - servomin) * bearing / 360);  

 
//Serial.print("ServoPosition: ");
//Serial.println(servoPosition);
  positionServo( &servoPointer, servoPosition );     // "sweepRange" = Max sweep; whole line gives position for servo

  // debug code below shows if compass and gps are giving correct info to servo
  // (may cause servo to stutter while active) 
   
 ///* 
   Serial.print("Heading: "); 
   Serial.print(g_currentHeading);
   Serial.println();
   Serial.print("Direction: ");
   Serial.print(directionToDestination); //g_directionToDestination
   Serial.println();
   Serial.print("Distance: ");
   Serial.print(distanceToDestination); //g_distanceToDestination 
   Serial.println();
   Serial.print("bearing: ");
   Serial.print(bearing);
   Serial.println();
   Serial.print("servo value: ");
   Serial.print(servoPosition);
   Serial.println();
   Serial.print("_______________");
   Serial.println();

 // */
} 

bool DataValid()
{
  return true;
}


